# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Device Details """

from abc import ABC
from dataclasses import dataclass


@dataclass
class DeviceDetails(ABC):
    """Generic Device Details class to include the Device representation"""

    _str = ""

    def __str__(self) -> str:
        """String representation of DeviceDetails

        Returns:
            str: String representationDeviceDetails
        """
        return self._str

    def __repr__(self) -> str:
        """String representation of DeviceDetails

        Returns:
            str: String representationDeviceDetails
        """
        return self._str
